#ifndef _IPT_PORTKNG_H
#define _IPT_PORTKNG_H

#define IPT_PORTKNG_KNOCKS 15

#define IPT_PORTKNG_MATCH_KNOCK				0x01
#define IPT_PORTKNG_MATCH_TIMEOUT			0x02
#define IPT_PORTKNG_MATCH_KNOCK_TIMEOUT	0x04

struct ipt_portkng_info {
	u_int8_t count;
	u_int16_t ports[IPT_PORTKNG_KNOCKS];
};


#endif /* _IPT_PORTKNG_H */
